/*
 * Name:    webportfwd
 * Desc:    program for forwarding http requests to ports on the local
 *          system to another URL
 * Utility: if you want http connections to different ports on a system to
 *          be redirected to different URLs
 * Author:  Edsel Adap <edsel@adap.org>
 * Copyright (c) 2002 by Edsel Adap.
 * License: GPL
 * History:
 *          Version 0.01		11/21/02 	EA
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <errno.h>
#include <limits.h>
#include <libgen.h>

#define FAILURE	-1
#define	SUCCESS 0
#define CONFIG_FILE	"webportfwd.conf"
#define TIMEOUT 300
#define	VERSION		"0.01"

typedef struct fwd_tbl {
	int port;
	char *url;
	struct fwd_tbl *next;
} Fwd_tbl;

int		readconf(char *filename);
int		startserver(int port, char *url);
void	strip(char *buf);
void	parse(char *buf, char *port, char *url);
void	add_entry(char *port, char *url);
void	do_work(int sockfd, char *url);
void	read_timeout(int a);
void	usage(void);
Fwd_tbl	*find_entry(Fwd_tbl *key);

Fwd_tbl *entries = NULL;
char *prog;

int
main(int argc, char **argv)
{
	Fwd_tbl *e;
	char filename[PATH_MAX];
	char c;
	extern char *optarg;
	extern int optind, optopt, opterr;

	opterr = 0;
	prog = basename(argv[0]);

	snprintf(filename, PATH_MAX, "%s/etc/%s", INSTALL_PREFIX, CONFIG_FILE);

	while((c = getopt(argc, argv, "c:hV")) != EOF) {
	    switch (c) {
		case 'c':
		    snprintf(filename, PATH_MAX, "%s", optarg);
		    break;
		case 'h':
		    usage();
		    exit(SUCCESS);
		    break;
		case 'V':
		    printf("%s version %s\nWritten by: Edsel Adap\n",
		        prog, VERSION);
		    exit(SUCCESS);
		    break;
		case '?':
		    usage();
		    exit(FAILURE);
		    break;
		default:
		    usage();
		    exit(SUCCESS);
		    break;
	    }
	}

	if (optind != argc) {
	    usage();
	    exit(FAILURE);
	}

	/* We don't want to create zombies */
	signal(SIGCHLD, SIG_IGN);

	/* Parse the config file and create the forwarding table */
	if(readconf(filename)) {
		fprintf(stderr,"Unable to process config file\n");
		exit (1);
	}

	/* Start servers for each entry in the table */
	e = entries;
	while(e){
		fprintf(stderr,"Starting service at port %d\n",e->port);
		if (startserver(e->port, e->url) == FAILURE) {
			fprintf(stderr,"Warning: failed to start service at port %d\n",
			    e->port);
		}
		e=e->next;
	}
}

/*
 * read the configuration file and initialize the forwarding table
 */
int readconf(char *filename){
	FILE *fp;
	char buf[BUFSIZ];
	char port[BUFSIZ], url[BUFSIZ];
	
	if ((fp = fopen(filename,"ro")) == NULL) {
		fprintf(stderr,"failed to open %s: %s\n",
			filename, strerror(errno));
		return(FAILURE);
	}
	while(fgets(buf, BUFSIZ, fp) != NULL) {
		strip(buf);
		if (buf[0] == '#')
			continue;
		parse(buf, port, url);
		add_entry(port,url);
	}
}

/*
 * remove leading spaces
 */
void
strip(char *buf)
{
	char *p, *q;

	p = q = buf;
	while(isspace(*p))
		p++;

	while(*p != '\0'){
		*q = *p;
		p++;
		q++;
	}
}

/*
 * parse the input into port and url fields
 */
void
parse(char *buf, char *port, char *url)
{
	char *p = buf;
	char *lp = port;
	char *ll = url;

	while(!isspace(*p)) {
		*lp=*p;
		lp++;
		p++;
	}
	*lp='\0';

	while(isspace(*p)){
		p++;
	}

	while(!isspace(*p)) {
		*ll=*p;
		ll++;
		p++;
	}
	*ll='\0';
}

/*
 * Add a new entry
 */
void
add_entry(char *port, char *url)
{
	Fwd_tbl *e, *s;

	e = malloc(sizeof(Fwd_tbl));
	if (e == NULL) {
		fprintf(stderr, "Error: memory allocation failed\n");
		exit(FAILURE);
	}

	e->port = atoi(port);
	e->url = strdup(url);

	if (entries == NULL) {
		entries = e;
		entries->next = NULL;
	} else {
		if (find_entry(e) == NULL) {
			e->next = entries;
			entries = e;
		} else {
			fprintf(stderr,"Duplicate entry for port %d found\n", e->port);
			exit(FAILURE);
		}
	}
}

/*
 * Search for an existing entry
 */
Fwd_tbl *
find_entry(Fwd_tbl *key)
{
	Fwd_tbl *t;

	t = entries;

	while(t){
		if (t->port == key->port)
			return(t);

		t=t->next;
	}
	return NULL;
}

/*
 * start server
 */
int
startserver(int port, char *url){
	int sockfd, newsockfd, clilen;
	struct sockaddr_in serv_addr, cli_addr;
	pid_t childpid, daemonpid;

	if ( (daemonpid = fork()) < 0 ) {
		fprintf(stderr,"fork error: %s\n", strerror(errno));
		return(FAILURE);
	}else if (daemonpid == 0) {
		/* we are the child */
		/* Open a TCP socket */
		if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
			fprintf(stderr,"Server: can't open stream socket\n");
			exit(1);
   		}

		/* Bind our local address so that the client can send to us */
		memset((char *) &serv_addr,0, sizeof(serv_addr));
		serv_addr.sin_family = AF_INET;
		serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
		serv_addr.sin_port = htons (port);
		if (bind(sockfd,(struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0){
			fprintf(stderr,"Server: can't bind local address");
			exit(FAILURE);
		}

		listen(sockfd,5);

		while(1){
			/* Wait for a connection from a client process. */
	
			clilen=sizeof(cli_addr);
			if ((newsockfd = accept(sockfd, (struct sockaddr *) &cli_addr,
			    & clilen)) < 0){
				fprintf(stderr,"server: accept error\n");
				exit(FAILURE);
			}

			if ( ( childpid = fork()) < 0) {
				fprintf(stderr,"server: fork error\n");
				exit(FAILURE);
			}else if (childpid == 0) {	/* child process */
				close(sockfd);			/* close original socked */
				do_work(newsockfd, url);
				exit(SUCCESS);
			}

			close(newsockfd);
		}
	}
	return(SUCCESS);
}

void
do_work(int sockfd, char *url)
{
	char mesg[BUFSIZ];
	int n;
	char *response="HTTP/1.1 302 Found\nLocation: ";

	/* setup timeout handler */
	signal(SIGALRM, read_timeout);

	alarm(TIMEOUT);
	
	while((n = read(sockfd, mesg, BUFSIZ)) > 0) {
		if (n > 0 ) {
			mesg[n]=0;
			if (strcmp(mesg+n-3,"\n\r\n") == 0){
				write(sockfd, response, strlen(response));
				write(sockfd, url, strlen(url));
				write(sockfd, "\n\n", 2);
				close(sockfd);
				return;
			}
		}
	}
}

/*
 * Usage Function
 */
void
usage()
{
	fprintf(stderr,"%s [-hV] [-c config_file]\n", prog);
}

/*
 * Timeout Signal handler
 */
void
read_timeout(int a)
{
	exit(FAILURE);
}

